﻿using System;
using System.Runtime.InteropServices;
using System.Text;

namespace RoboBrain
{
    static class IniClass
    {
        public static string FileName;
        public static string Section;

        [DllImport("kernel32.dll", EntryPoint = "GetPrivateProfileStringW", CharSet = CharSet.Unicode, SetLastError = true, ExactSpelling = true)]
        private static extern int GetPrivateProfileString(string lpApplicationName, string lpKeyName, string lpDefault, string lpValue, int nSize, string lpFileName);
        [DllImport("kernel32.dll", EntryPoint = "WritePrivateProfileStringW", CharSet = CharSet.Unicode, SetLastError = true, ExactSpelling = true)]
        private static extern int WritePrivateProfileString(string lpApplicationName, string lpKeyName, string lpString, string lpFileName);
        [DllImport("kernel32.dll", EntryPoint = "GetPrivateProfileIntW", CharSet = CharSet.Unicode, SetLastError = true, ExactSpelling = true)]
        private static extern int GetPrivateProfileInt(string lpApplicationName, string lpKeyName, int nDefault, string lpFileName);
        [DllImport("kernel32.dll", EntryPoint = "WritePrivateProfileStringW", CharSet = CharSet.Unicode, SetLastError = true, ExactSpelling = true)]
        private static extern int FlushPrivateProfileString(int lpApplicationName, int lpKeyName, int lpString, string lpFileName);


        public static void Flush()
        {
            FlushPrivateProfileString(0, 0, 0, FileName);
        }

        public static void Write(string Key, string Value)
        {
            WritePrivateProfileString(Section, Key, Value, FileName);
        }

        public static void Write(string Key, int Value)
        {
            WritePrivateProfileString(Section, Key, Value.ToString(), FileName);
        }

        public static void Write(string Key, bool Value)
        {
            WritePrivateProfileString(Section, Key, (Value ? "1" : "0"), FileName);
        }

        public static string Read(string Key, string Default)
        {
            string ret = new string(' ', 256);
            int i = GetPrivateProfileString(Section, Key, Default, ret, 256, FileName);
            return ret.Substring(0, i);
        }

        public static int Read(string Key, int Default)
        {
            return GetPrivateProfileInt(Section, Key, Default, FileName);
        }

        public static bool Read(string Key, bool Default)
        {
            return GetPrivateProfileInt(Section, Key, (Default ? 1 : 0), FileName) == 1;
        }

    }
}
