﻿using System;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using Emgu.CV;
using Emgu.CV.CvEnum;
using Emgu.CV.Structure;
using AForge.Video;


namespace RoboBrain
{
    public partial class mainForm : Form
    {
        [DllImport("user32.dll")] static extern uint GetAsyncKeyState(int vKey);

        private RobotClass robot;
        private scCommand[] cmds;

        MJPEGStream video_stream;
        Boolean working = false; // true while working with a frame
        private Emgu.CV.Image<Bgr, Byte> frame; // last captured frame
        HaarCascade face_cascade = null;
        MCvSURFParams surfParam;
        Image<Gray, Byte> marker;
        SURFFeature[] markerFeatures;
        
        private string OpenFileName = "";
        private bool Playing = false;
        private int CommandIndex = -1;
        private uint WaitUntil = 0;
        

        #region Main Form

        public mainForm()
        {
            InitializeComponent();

            cmds = new scCommand[] {
                scCommand.DIGITAL_WRITE,
                scCommand.DIGITAL_READ,
                scCommand.ANALOG_READ,
                scCommand.PIN_MODE,

                scCommand.SERVO_SET_POSITION,
                scCommand.SERVO_GET_POSITION,
                scCommand.SERVO_SET_SPEED,
                scCommand.SERVO_GET_SPEED,
                scCommand.SERVO_SET_TIME,
                scCommand.SERVO_GET_TIME,
                scCommand.SERVO_STOP,
                scCommand.SERVO_DISABLE,
                scCommand.SERVO_MOVING,

                scCommand.SERVO_SET_PIN,
                scCommand.SERVO_GET_PIN,
                scCommand.SERVO_SET_ZERO,
                scCommand.SERVO_GET_ZERO,
                scCommand.SERVO_INIT,
                scCommand.SERVO_DETACH,
                
                scCommand.SERVO_GROUP_SPEED,
                scCommand.SERVO_GROUP_TIME,
                //scCommand.SERVO_GROUP_MOVING,
                
                scCommand.RESET_GAMEPAD,
                scCommand.UPTIME,
                scCommand.CHECK};
            
            cboCommand.Items.Add("DIGITAL_WRITE");
            cboCommand.Items.Add("DIGITAL_READ");
            cboCommand.Items.Add("ANALOG_READ");
            cboCommand.Items.Add("PIN_MODE");

            cboCommand.Items.Add("SERVO_SET_POSITION");
            cboCommand.Items.Add("SERVO_GET_POSITION");
            cboCommand.Items.Add("SERVO_SET_SPEED");
            cboCommand.Items.Add("SERVO_GET_SPEED");
            cboCommand.Items.Add("SERVO_SET_TIME");
            cboCommand.Items.Add("SERVO_GET_TIME");
            cboCommand.Items.Add("SERVO_STOP");
            cboCommand.Items.Add("SERVO_DISABLE");
            cboCommand.Items.Add("SERVO_MOVING");
            
            cboCommand.Items.Add("SERVO_SET_PIN");
            cboCommand.Items.Add("SERVO_GET_PIN");
            cboCommand.Items.Add("SERVO_SET_ZERO");
            cboCommand.Items.Add("SERVO_GET_ZERO");
            cboCommand.Items.Add("SERVO_INIT");
            cboCommand.Items.Add("SERVO_DETACH");
            
            cboCommand.Items.Add("SERVO_GROUP_SPEED");
            cboCommand.Items.Add("SERVO_GROUP_TIME");
            //cboCommand.Items.Add("SERVO_GROUP_MOVING");

            cboCommand.Items.Add("RESET_GAMEPAD");
            cboCommand.Items.Add("UPTIME");
            cboCommand.Items.Add("CHECK");
            
            Transport.OnMessage += new Transport.OnMessageHandler(ComPort_OnMessage);
            robot = new RobotClass();
            ReadConfig();

            UpdateLayout();
        }

        private void btnExit_Click(object sender, EventArgs e)
        {
            try
            {
                btnStop.PerformClick(); // stop playback
                robot.EnableAllServos(false); // disable all
                Transport.Close(); // close COM port
                this.Close();
            }
            catch (Exception ex)
            {
                log("Ошибка. " + ex.Message);
            }
        }

        private void mainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            DisableCamera();
        }

        private static bool IsKeyDown(System.Windows.Forms.Keys vKey)
        {
            return 0 != (GetAsyncKeyState((int)vKey) & 0x8000);
        }

        private void btnSaveConfig_Click(object sender, EventArgs e)
        {
            try
            {
                IniClass.FileName = Application.ExecutablePath.Substring(0, Application.ExecutablePath.Length - 3) + "ini";
                IniClass.Section = "General";
                IniClass.Write("Script", ShortenPath(OpenFileName));
                IniClass.Write("Port", tbPort.Text);
                IniClass.Write("DriveSpeed", udDriveSpeed.Value.ToString());
                IniClass.Write("ServoSpeed", udServoSpeed.Value.ToString());
                IniClass.Write("ServoStep", udServoStep.Value.ToString());
                IniClass.Write("SmoothSpeedControl", chbSmoothAcceleration.Checked ? 1 : 0);
                IniClass.Write("FollowFace", chbFollowFace.Checked ? 1 : 0);
                IniClass.Write("Layout", btnLayout.Checked ? 1 : 0);
                IniClass.Write("Maximized", this.WindowState == FormWindowState.Maximized ? 1 : 0);
                if (!btnLayout.Checked && this.WindowState != FormWindowState.Maximized)
                {
                    IniClass.Write("Left", this.Left);
                    IniClass.Write("Top", this.Top);
                    IniClass.Write("Width", this.ClientSize.Width);
                    IniClass.Write("Height", this.ClientSize.Height);
                }
                robot.SaveToIni();
                IniClass.Flush();
                log("config saved (" + IniClass.FileName + ")");
            }
            catch (Exception ex)
            {
                log("Ошибка при сохранении конфигурации. " + ex.Message);
            }
        }

        private void ReadConfig()
        {
            string s = "";

            try
            {
                IniClass.FileName = Application.ExecutablePath.Substring(0, Application.ExecutablePath.Length - 3) + "ini";
                IniClass.Section = "General";
                s = ExpandPath(IniClass.Read("Script", ""));
                tbPort.Text = IniClass.Read("Port", "COM1");
                udDriveSpeed.Value = Int32.Parse(IniClass.Read("DriveSpeed", "20"));
                udServoSpeed.Value = Int32.Parse(IniClass.Read("ServoSpeed", "1000"));
                udServoStep.Value = Int32.Parse(IniClass.Read("ServoStep", "2"));
                chbSmoothAcceleration.Checked = IniClass.Read("SmoothSpeedControl", "0") == "1";
                chbFollowFace.Checked = IniClass.Read("FollowFace", "0") == "1";
                btnLayout.Checked = IniClass.Read("Layout", "0") == "1";
                this.ResumeLayout();
                if (IniClass.Read("Maximized", "1") == "1") this.WindowState = FormWindowState.Maximized;
                else this.WindowState = FormWindowState.Normal;
                this.ResumeLayout();
                if (!btnLayout.Checked && this.WindowState != FormWindowState.Maximized)
                {
                    this.Left = Int32.Parse(IniClass.Read("Left", "20"));
                    this.Top = Int32.Parse(IniClass.Read("Top", "20"));
                    this.SetClientSizeCore(Int32.Parse(IniClass.Read("Width", "800")), Int32.Parse(IniClass.Read("Height", "600")));
                }
                robot.ReadFromIni();
                log("config loaded (" + IniClass.FileName + ")");
            }
            catch (Exception ex)
            {
                log("Ошибка при загрузке конфигурации. " + ex.Message);
            }

            LoadScript(s);
            ShowServoParams();
        }

        public string ShortenPath(string path)
        {
            string pp = System.IO.Path.GetDirectoryName(Application.ExecutablePath);
            return path.Replace(pp, "%pp%");
        }

        public string ExpandPath(string path)
        {
            string pp = System.IO.Path.GetDirectoryName(Application.ExecutablePath);
            return path.Replace("%pp%", pp);
        }

        public void log(string str) // thread-safe method
        {
            if (this.rtLog.InvokeRequired)
            {
                this.rtLog.BeginInvoke(new MethodInvoker(delegate() { this.rtLog.AppendText(str + "\n"); }));
                this.rtLog.BeginInvoke(new MethodInvoker(delegate() { this.rtLog.Update(); }));
            }
            else
            {
                rtLog.AppendText(str + "\n");
                rtLog.Update();
            }
        }

        public void addLog(string str) // thread-safe method
        {
            if (this.rtLog.InvokeRequired)
            {
                this.rtLog.BeginInvoke(new MethodInvoker(delegate() { this.rtLog.AppendText(str); }));
                this.rtLog.BeginInvoke(new MethodInvoker(delegate() { this.rtLog.Update(); }));
            }
            else
            {
                rtLog.AppendText(str);
                rtLog.Update();
            }
        }

        private void ComPort_OnMessage(object o, string message)
        {
            log(message);
        }

        private void btnConnect_Click(object sender, EventArgs e)
        {
            try
            {
                if (Transport.active)
                {
                    addLog("Закрываю порт ... ");
                    Transport.Close();
                    if (!Transport.active) log("готово");
                    else log("не удалось выполнить");
                }
                else
                {
                    addLog("Открываю порт ... ");
                    Transport.Open(tbPort.Text);
                    if (Transport.active)
                    {
                        log("готово");
                        robot.CheckConnection(); // check if connection successful
                        robot.ReadPins(); // check if connection successful
                        robot.SetServoSpeed(Decimal.ToInt32(udServoSpeed.Value));
                    }
                    else log("не удалось выполнить");
                }
            }
            catch (Exception ex)
            {
                log("Ошибка при открытии порта. " + ex.Message);
            }
        }

        private void btnbtnKeyboardControl_Click(object sender, EventArgs e)
        {
            groupRobot.Enabled = !btnKeyboardControl.Checked;
            groupControls.Enabled = !btnKeyboardControl.Checked;
            tabs.Enabled = !btnKeyboardControl.Checked;
            UpdateInfo();
        }

        private void btnServosToClipboard_Click(object sender, EventArgs e)
        {
            Clipboard.Clear();
            string str = robot.ToStringD();
            if (str.Length > 0) Clipboard.SetText(str); // clipboard does not accept empty strings
        }

        private void btnDisableAllServos_Click(object sender, EventArgs e)
        {
            robot.EnableAllServos(false);
        }

        private void udServoSpeed_ValueChanged(object sender, EventArgs e)
        {
            robot.SetServoSpeed(Decimal.ToInt32(udServoSpeed.Value));
        }

        private void btnSend_Click(object sender, EventArgs e)
        {
            if (tbParam1.Text == "") tbParam1.Text = "0";
            if (tbParam2.Text == "") tbParam2.Text = "0";
            if (cboCommand.SelectedIndex < 0) cboCommand.SelectedIndex = cboCommand.Items.Count - 1;
            scCommand cmd = cmds[cboCommand.SelectedIndex];
            Transport.Cmd(cmd, int.Parse(tbParam1.Text), int.Parse(tbParam2.Text));
        }

        private void chbLight1_CheckedChanged(object sender, EventArgs e)
        {
            robot.DigitalWrite(52, chbLight1.Checked ? 1 : 0);
        }

        private void chbLight2_CheckedChanged(object sender, EventArgs e)
        {
            robot.DigitalWrite(53, chbLight2.Checked ? 1 : 0);
        }

        private void btnLayout_Click(object sender, EventArgs e)
        {
            btnLayout.Checked = !btnLayout.Checked;
            UpdateLayout();
        }

        private void btnClearTerminal_Click(object sender, EventArgs e)
        {
            rtLog.Clear();
        }

        private void timer_Tick(object sender, EventArgs e)
        {
            if (btnKeyboardControl.Checked) ReadKeys();
            if (Playing) DoCommand();
            UpdateInfo();
        }

        private void chbSmoothSpeedControl_CheckedChanged(object sender, EventArgs e)
        {
            if (chbSmoothAcceleration.Checked)
            {
                Transport.Cmd(scCommand.SERVO_SET_SPEED, 0, Decimal.ToInt32(udServoSpeed.Value));
                Transport.Cmd(scCommand.SERVO_SET_SPEED, 1, Decimal.ToInt32(udServoSpeed.Value));
            }
            else
            {
                Transport.Cmd(scCommand.SERVO_SET_SPEED, 0, 0);
                Transport.Cmd(scCommand.SERVO_SET_SPEED, 1, 0);
            }
        }

        private void chbLogData_CheckedChanged(object sender, EventArgs e)
        {
            Transport.logData = chbLogDetails.Checked;
        }

        private void statusTimer_Tick(object sender, EventArgs e)
        {
            Transport.Cmd(scCommand.ANALOG_READ, 8, 0);
            //Transport.Cmd(scCommand.ANALOG_READ, 9, 0);
            Transport.Cmd(scCommand.UPTIME, 0, 0);
        }

        private void btnShowServoParams_Click(object sender, EventArgs e)
        {
            robot.ReadPins();
            ShowServoParams();
        }

        private void btnApplyServoParams_Click(object sender, EventArgs e)
        {
            robot.Servos[0].MinPosition = int.Parse(tb0min.Text);
            robot.Servos[1].MinPosition = int.Parse(tb1min.Text);
            robot.Servos[2].MinPosition = int.Parse(tb2min.Text);
            robot.Servos[3].MinPosition = int.Parse(tb3min.Text);
            robot.Servos[4].MinPosition = int.Parse(tb4min.Text);
            robot.Servos[5].MinPosition = int.Parse(tb5min.Text);
            robot.Servos[6].MinPosition = int.Parse(tb6min.Text);
            robot.Servos[7].MinPosition = int.Parse(tb7min.Text);
            robot.Servos[8].MinPosition = int.Parse(tb8min.Text);
            robot.Servos[9].MinPosition = int.Parse(tb9min.Text);
            robot.Servos[10].MinPosition = int.Parse(tb10min.Text);
            robot.Servos[11].MinPosition = int.Parse(tb11min.Text);
            robot.Servos[12].MinPosition = int.Parse(tb12min.Text);
            robot.Servos[13].MinPosition = int.Parse(tb13min.Text);
            robot.Servos[14].MinPosition = int.Parse(tb14min.Text);

            robot.Servos[0].NullPosition = int.Parse(tb0null.Text);
            robot.Servos[1].NullPosition = int.Parse(tb1null.Text);
            robot.Servos[2].NullPosition = int.Parse(tb2null.Text);
            robot.Servos[3].NullPosition = int.Parse(tb3null.Text);
            robot.Servos[4].NullPosition = int.Parse(tb4null.Text);
            robot.Servos[5].NullPosition = int.Parse(tb5null.Text);
            robot.Servos[6].NullPosition = int.Parse(tb6null.Text);
            robot.Servos[7].NullPosition = int.Parse(tb7null.Text);
            robot.Servos[8].NullPosition = int.Parse(tb8null.Text);
            robot.Servos[9].NullPosition = int.Parse(tb9null.Text);
            robot.Servos[10].NullPosition = int.Parse(tb10null.Text);
            robot.Servos[11].NullPosition = int.Parse(tb11null.Text);
            robot.Servos[12].NullPosition = int.Parse(tb12null.Text);
            robot.Servos[13].NullPosition = int.Parse(tb13null.Text);
            robot.Servos[14].NullPosition = int.Parse(tb14null.Text);

            robot.Servos[0].MaxPosition = int.Parse(tb0max.Text);
            robot.Servos[1].MaxPosition = int.Parse(tb1max.Text);
            robot.Servos[2].MaxPosition = int.Parse(tb2max.Text);
            robot.Servos[3].MaxPosition = int.Parse(tb3max.Text);
            robot.Servos[4].MaxPosition = int.Parse(tb4max.Text);
            robot.Servos[5].MaxPosition = int.Parse(tb5max.Text);
            robot.Servos[6].MaxPosition = int.Parse(tb6max.Text);
            robot.Servos[7].MaxPosition = int.Parse(tb7max.Text);
            robot.Servos[8].MaxPosition = int.Parse(tb8max.Text);
            robot.Servos[9].MaxPosition = int.Parse(tb9max.Text);
            robot.Servos[10].MaxPosition = int.Parse(tb10max.Text);
            robot.Servos[11].MaxPosition = int.Parse(tb11max.Text);
            robot.Servos[12].MaxPosition = int.Parse(tb12max.Text);
            robot.Servos[13].MaxPosition = int.Parse(tb13max.Text);
            robot.Servos[14].MaxPosition = int.Parse(tb14max.Text);

            robot.Servos[0].Inverse = chb0inv.Checked;
            robot.Servos[1].Inverse = chb1inv.Checked;
            robot.Servos[2].Inverse = chb2inv.Checked;
            robot.Servos[3].Inverse = chb3inv.Checked;
            robot.Servos[4].Inverse = chb4inv.Checked;
            robot.Servos[5].Inverse = chb5inv.Checked;
            robot.Servos[6].Inverse = chb6inv.Checked;
            robot.Servos[7].Inverse = chb7inv.Checked;
            robot.Servos[8].Inverse = chb8inv.Checked;
            robot.Servos[9].Inverse = chb9inv.Checked;
            robot.Servos[10].Inverse = chb10inv.Checked;
            robot.Servos[11].Inverse = chb11inv.Checked;
            robot.Servos[12].Inverse = chb12inv.Checked;
            robot.Servos[13].Inverse = chb13inv.Checked;
            robot.Servos[14].Inverse = chb14inv.Checked;

            robot.Servos[0].Position = int.Parse(tb0pos.Text);
            robot.Servos[1].Position = int.Parse(tb1pos.Text);
            robot.Servos[2].Position = int.Parse(tb2pos.Text);
            robot.Servos[3].Position = int.Parse(tb3pos.Text);
            robot.Servos[4].Position = int.Parse(tb4pos.Text);
            robot.Servos[5].Position = int.Parse(tb5pos.Text);
            robot.Servos[6].Position = int.Parse(tb6pos.Text);
            robot.Servos[7].Position = int.Parse(tb7pos.Text);
            robot.Servos[8].Position = int.Parse(tb8pos.Text);
            robot.Servos[9].Position = int.Parse(tb9pos.Text);
            robot.Servos[10].Position = int.Parse(tb10pos.Text);
            robot.Servos[11].Position = int.Parse(tb11pos.Text);
            robot.Servos[12].Position = int.Parse(tb12pos.Text);
            robot.Servos[13].Position = int.Parse(tb13pos.Text);
            robot.Servos[14].Position = int.Parse(tb14pos.Text);
        }

        private void btnWritePins_Click(object sender, EventArgs e)
        {
            if (DialogResult.Yes == MessageBox.Show("Confirm pins change", "Attention", MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation, MessageBoxDefaultButton.Button2))
            try
            {
                Transport.Cmd(scCommand.SERVO_SET_PIN, 0, int.Parse(tb0pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 1, int.Parse(tb1pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 2, int.Parse(tb2pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 3, int.Parse(tb3pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 4, int.Parse(tb4pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 5, int.Parse(tb5pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 6, int.Parse(tb6pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 7, int.Parse(tb7pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 8, int.Parse(tb8pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 9, int.Parse(tb9pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 10, int.Parse(tb10pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 11, int.Parse(tb11pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 12, int.Parse(tb12pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 13, int.Parse(tb13pin.Text));
                Transport.Cmd(scCommand.SERVO_SET_PIN, 14, int.Parse(tb14pin.Text));
                Transport.Cmd(scCommand.SERVO_INIT, 0, 0);
            }
            catch (Exception ex)
            {
                log("error writing servo pins. " + ex.Message);
            }
        }

        private void UpdateLayout()
        {
            split.IsSplitterFixed = btnLayout.Checked;
            if (btnLayout.Checked)
            {
                this.WindowState = FormWindowState.Normal;
                this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.FixedSingle;
                this.Width = split.SplitterDistance + 14;
                this.Height = groupControls.Bottom + 72;
            }
            else
            {
                this.FormBorderStyle = System.Windows.Forms.FormBorderStyle.Sizable;
                this.WindowState = FormWindowState.Maximized;
            }
        }

        private void UpdateInfo()
        {
            Color EnabledColor = Color.FromArgb(200, 255, 180);
            Color DefaultColor = this.BackColor;
            Color WarningColor = Color.FromArgb(255, 255, 200);
            Color ErrorColor = Color.Red;

            lblConnected.Text = Transport.connected ? "Connected" : "Not connected";
            lblConnected.BackColor = Transport.connected ? EnabledColor : DefaultColor;

            lblBattery1.Text = "Bat 1: " + (Transport.analog[8] == 0 ? "--" : (Transport.analog[8] * 4).ToString("0.00") + "V");
            lblBattery2.Text = "Bat 2: " + (Transport.analog[9] == 0 ? "--" : (Transport.analog[9] * 2).ToString("0.00") + "V");
            
            if (!Transport.connected || Transport.analog[8] == 0) lblBattery1.BackColor = DefaultColor;
            else if (Transport.analog[8] < 3) lblBattery1.BackColor = ErrorColor;
            else if (Transport.analog[8] < 3.25) lblBattery1.BackColor = WarningColor;
            else lblBattery1.BackColor = DefaultColor;

            if (!Transport.connected || Transport.analog[9] == 0) lblBattery2.BackColor = DefaultColor;
            else if (Transport.analog[9] < 3) lblBattery2.BackColor = ErrorColor;
            else if (Transport.analog[9] < 3.25) lblBattery2.BackColor = WarningColor;
            else lblBattery2.BackColor = DefaultColor;

            if (Transport.uptime == 0) lblUptime.Text = "Up: --";
            else
            {
                uint h, m, s;
                h = Convert.ToUInt32(Transport.uptime / 3600000);
                m = Convert.ToUInt32(Transport.uptime / 60000) - h * 60;
                s = Convert.ToUInt32(Transport.uptime / 1000) - h * 3600 - m * 60;
                lblUptime.Text = "Up: " + (h < 10 ? "0" + h.ToString() : h.ToString()) + ":" +
                    (m < 10 ? "0" + m.ToString() : m.ToString()) + ":" +
                    (s < 10 ? "0" + s.ToString() : s.ToString());
            }
            
            this.Text = Playing ? "RoboBrain - Playing" : "RoboBrain";

            tb0.Text = robot.GetServoAngle("r_track").ToString();
            tb1.Text = robot.GetServoAngle("l_track").ToString();
            tb2.Text = robot.GetServoAngle("base").ToString();
            tb3.Text = robot.GetServoAngle("r_shoulder_rot").ToString();
            tb4.Text = robot.GetServoAngle("r_shoulder_ud").ToString();
            tb5.Text = robot.GetServoAngle("r_elbow").ToString();
            tb6.Text = robot.GetServoAngle("r_wrist").ToString();
            tb7.Text = robot.GetServoAngle("r_hand").ToString();
            tb8.Text = robot.GetServoAngle("l_shoulder_rot").ToString();
            tb9.Text = robot.GetServoAngle("l_shoulder_ud").ToString();
            tb10.Text = robot.GetServoAngle("l_elbow").ToString();
            tb11.Text = robot.GetServoAngle("l_wrist").ToString();
            tb12.Text = robot.GetServoAngle("l_hand").ToString();
            tb13.Text = robot.GetServoAngle("head_pan").ToString();
            tb14.Text = robot.GetServoAngle("head_tilt").ToString();

            tb0.BackColor = robot.GetServoEnabled("r_track") ? EnabledColor : DefaultColor;
            tb1.BackColor = robot.GetServoEnabled("l_track") ? EnabledColor : DefaultColor;
            tb2.BackColor = robot.GetServoEnabled("base") ? EnabledColor : DefaultColor;
            tb3.BackColor = robot.GetServoEnabled("r_shoulder_rot") ? EnabledColor : DefaultColor;
            tb4.BackColor = robot.GetServoEnabled("r_shoulder_ud") ? EnabledColor : DefaultColor;
            tb5.BackColor = robot.GetServoEnabled("r_elbow") ? EnabledColor : DefaultColor;
            tb6.BackColor = robot.GetServoEnabled("r_wrist") ? EnabledColor : DefaultColor;
            tb7.BackColor = robot.GetServoEnabled("r_hand") ? EnabledColor : DefaultColor;
            tb8.BackColor = robot.GetServoEnabled("l_shoulder_rot") ? EnabledColor : DefaultColor;
            tb9.BackColor = robot.GetServoEnabled("l_shoulder_ud") ? EnabledColor : DefaultColor;
            tb10.BackColor = robot.GetServoEnabled("l_elbow") ? EnabledColor : DefaultColor;
            tb11.BackColor = robot.GetServoEnabled("l_wrist") ? EnabledColor : DefaultColor;
            tb12.BackColor = robot.GetServoEnabled("l_hand") ? EnabledColor : DefaultColor;
            tb13.BackColor = robot.GetServoEnabled("head_pan") ? EnabledColor : DefaultColor;
            tb14.BackColor = robot.GetServoEnabled("head_tilt") ? EnabledColor : DefaultColor;

            tb0pos.Text = robot.Servos[0].Position.ToString();
            tb1pos.Text = robot.Servos[1].Position.ToString();
            tb2pos.Text = robot.Servos[2].Position.ToString();
            tb3pos.Text = robot.Servos[3].Position.ToString();
            tb4pos.Text = robot.Servos[4].Position.ToString();
            tb5pos.Text = robot.Servos[5].Position.ToString();
            tb6pos.Text = robot.Servos[6].Position.ToString();
            tb7pos.Text = robot.Servos[7].Position.ToString();
            tb8pos.Text = robot.Servos[8].Position.ToString();
            tb9pos.Text = robot.Servos[9].Position.ToString();
            tb10pos.Text = robot.Servos[10].Position.ToString();
            tb11pos.Text = robot.Servos[11].Position.ToString();
            tb12pos.Text = robot.Servos[12].Position.ToString();
            tb13pos.Text = robot.Servos[13].Position.ToString();
            tb14pos.Text = robot.Servos[14].Position.ToString();
        }

        private void ShowServoParams()
        {
            tb0name.Text = robot.Servos[0].Name;
            tb1name.Text = robot.Servos[1].Name;
            tb2name.Text = robot.Servos[2].Name;
            tb3name.Text = robot.Servos[3].Name;
            tb4name.Text = robot.Servos[4].Name;
            tb5name.Text = robot.Servos[5].Name;
            tb6name.Text = robot.Servos[6].Name;
            tb7name.Text = robot.Servos[7].Name;
            tb8name.Text = robot.Servos[8].Name;
            tb9name.Text = robot.Servos[9].Name;
            tb10name.Text = robot.Servos[10].Name;
            tb11name.Text = robot.Servos[11].Name;
            tb12name.Text = robot.Servos[12].Name;
            tb13name.Text = robot.Servos[13].Name;
            tb14name.Text = robot.Servos[14].Name;

            tb0pos.Text = robot.Servos[0].Position.ToString();
            tb1pos.Text = robot.Servos[1].Position.ToString();
            tb2pos.Text = robot.Servos[2].Position.ToString();
            tb3pos.Text = robot.Servos[3].Position.ToString();
            tb4pos.Text = robot.Servos[4].Position.ToString();
            tb5pos.Text = robot.Servos[5].Position.ToString();
            tb6pos.Text = robot.Servos[6].Position.ToString();
            tb7pos.Text = robot.Servos[7].Position.ToString();
            tb8pos.Text = robot.Servos[8].Position.ToString();
            tb9pos.Text = robot.Servos[9].Position.ToString();
            tb10pos.Text = robot.Servos[10].Position.ToString();
            tb11pos.Text = robot.Servos[11].Position.ToString();
            tb12pos.Text = robot.Servos[12].Position.ToString();
            tb13pos.Text = robot.Servos[13].Position.ToString();
            tb14pos.Text = robot.Servos[14].Position.ToString();

            tb0min.Text = robot.Servos[0].MinPosition.ToString();
            tb1min.Text = robot.Servos[1].MinPosition.ToString();
            tb2min.Text = robot.Servos[2].MinPosition.ToString();
            tb3min.Text = robot.Servos[3].MinPosition.ToString();
            tb4min.Text = robot.Servos[4].MinPosition.ToString();
            tb5min.Text = robot.Servos[5].MinPosition.ToString();
            tb6min.Text = robot.Servos[6].MinPosition.ToString();
            tb7min.Text = robot.Servos[7].MinPosition.ToString();
            tb8min.Text = robot.Servos[8].MinPosition.ToString();
            tb9min.Text = robot.Servos[9].MinPosition.ToString();
            tb10min.Text = robot.Servos[10].MinPosition.ToString();
            tb11min.Text = robot.Servos[11].MinPosition.ToString();
            tb12min.Text = robot.Servos[12].MinPosition.ToString();
            tb13min.Text = robot.Servos[13].MinPosition.ToString();
            tb14min.Text = robot.Servos[14].MinPosition.ToString();

            tb0null.Text = robot.Servos[0].NullPosition.ToString();
            tb1null.Text = robot.Servos[1].NullPosition.ToString();
            tb2null.Text = robot.Servos[2].NullPosition.ToString();
            tb3null.Text = robot.Servos[3].NullPosition.ToString();
            tb4null.Text = robot.Servos[4].NullPosition.ToString();
            tb5null.Text = robot.Servos[5].NullPosition.ToString();
            tb6null.Text = robot.Servos[6].NullPosition.ToString();
            tb7null.Text = robot.Servos[7].NullPosition.ToString();
            tb8null.Text = robot.Servos[8].NullPosition.ToString();
            tb9null.Text = robot.Servos[9].NullPosition.ToString();
            tb10null.Text = robot.Servos[10].NullPosition.ToString();
            tb11null.Text = robot.Servos[11].NullPosition.ToString();
            tb12null.Text = robot.Servos[12].NullPosition.ToString();
            tb13null.Text = robot.Servos[13].NullPosition.ToString();
            tb14null.Text = robot.Servos[14].NullPosition.ToString();

            tb0max.Text = robot.Servos[0].MaxPosition.ToString();
            tb1max.Text = robot.Servos[1].MaxPosition.ToString();
            tb2max.Text = robot.Servos[2].MaxPosition.ToString();
            tb3max.Text = robot.Servos[3].MaxPosition.ToString();
            tb4max.Text = robot.Servos[4].MaxPosition.ToString();
            tb5max.Text = robot.Servos[5].MaxPosition.ToString();
            tb6max.Text = robot.Servos[6].MaxPosition.ToString();
            tb7max.Text = robot.Servos[7].MaxPosition.ToString();
            tb8max.Text = robot.Servos[8].MaxPosition.ToString();
            tb9max.Text = robot.Servos[9].MaxPosition.ToString();
            tb10max.Text = robot.Servos[10].MaxPosition.ToString();
            tb11max.Text = robot.Servos[11].MaxPosition.ToString();
            tb12max.Text = robot.Servos[12].MaxPosition.ToString();
            tb13max.Text = robot.Servos[13].MaxPosition.ToString();
            tb14max.Text = robot.Servos[14].MaxPosition.ToString();

            chb0inv.Checked = robot.Servos[0].Inverse;
            chb1inv.Checked = robot.Servos[1].Inverse;
            chb2inv.Checked = robot.Servos[2].Inverse;
            chb3inv.Checked = robot.Servos[3].Inverse;
            chb4inv.Checked = robot.Servos[4].Inverse;
            chb5inv.Checked = robot.Servos[5].Inverse;
            chb6inv.Checked = robot.Servos[6].Inverse;
            chb7inv.Checked = robot.Servos[7].Inverse;
            chb8inv.Checked = robot.Servos[8].Inverse;
            chb9inv.Checked = robot.Servos[9].Inverse;
            chb10inv.Checked = robot.Servos[10].Inverse;
            chb11inv.Checked = robot.Servos[11].Inverse;
            chb12inv.Checked = robot.Servos[12].Inverse;
            chb13inv.Checked = robot.Servos[13].Inverse;
            chb14inv.Checked = robot.Servos[14].Inverse;

            tb0pin.Text = Transport.servo_pin[0].ToString();
            tb1pin.Text = Transport.servo_pin[1].ToString();
            tb2pin.Text = Transport.servo_pin[2].ToString();
            tb3pin.Text = Transport.servo_pin[3].ToString();
            tb4pin.Text = Transport.servo_pin[4].ToString();
            tb5pin.Text = Transport.servo_pin[5].ToString();
            tb6pin.Text = Transport.servo_pin[6].ToString();
            tb7pin.Text = Transport.servo_pin[7].ToString();
            tb8pin.Text = Transport.servo_pin[8].ToString();
            tb9pin.Text = Transport.servo_pin[9].ToString();
            tb10pin.Text = Transport.servo_pin[10].ToString();
            tb11pin.Text = Transport.servo_pin[11].ToString();
            tb12pin.Text = Transport.servo_pin[12].ToString();
            tb13pin.Text = Transport.servo_pin[13].ToString();
            tb14pin.Text = Transport.servo_pin[14].ToString();
        }

        #endregion


        #region Read keyboard

        /// <summary>
        /// hotkeys
        /// </summary>
        private void mainForm_KeyDown(object sender, KeyEventArgs e)
        {
            if (!e.Alt && !e.Control && !e.Shift)
            {
                if (e.KeyCode == Keys.F2)
                {
                    btnSaveScript.PerformClick();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.F3)
                {
                    btnOpenScript.PerformClick();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.F4)
                {
                    btnKeyboardControl.PerformClick();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.F5)
                {
                    btnConnect.PerformClick();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.F6)
                {
                    btnCamera.PerformClick();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.F7)
                {
                    tbSaveFrame.PerformClick();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.F8)
                {
                    btnLayout.PerformClick();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.F9)
                {
                    btnPlayPause.PerformClick();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.F10)
                {
                    btnStop.PerformClick();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.F11)
                {
                    btnStepPrevious.PerformClick();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.F12)
                {
                    btnStepNext.PerformClick();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.Escape)
                {
                    btnStop.PerformClick();
                    robot.EnableAllServos(false);
                }
            }
            else if (e.KeyCode == Keys.S && e.Control && !e.Alt && !e.Shift)
            {
                btnServosToClipboard.PerformClick();
                e.Handled = true;
            }
            else if (e.KeyCode == Keys.F4 && e.Alt && !e.Control && !e.Shift)
            {
                btnExit.PerformClick();
                e.Handled = true;
            }
            
            if (btnKeyboardControl.Checked)
            {
                if (e.KeyCode == Keys.NumPad0)
                {
                    chbLight1.Checked = !chbLight1.Checked;
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.Decimal)
                {
                    chbLight2.Checked = !chbLight2.Checked;
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.Multiply)
                {
                    chbSmoothAcceleration.Checked = !chbSmoothAcceleration.Checked;
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.Add)
                {
                    udDriveSpeed.UpButton();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.Subtract)
                {
                    udDriveSpeed.DownButton();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.PageUp)
                {
                    udServoSpeed.UpButton();
                    e.Handled = true;
                }
                else if (e.KeyCode == Keys.PageDown)
                {
                    udServoSpeed.DownButton();
                    e.Handled = true;
                }
            }
        }

        /// <summary>
        /// Asyncronous key reading
        /// </summary>
        private void ReadKeys()
        {
            int left = 0;
            int right = 0;
            int speed = Decimal.ToInt32(udDriveSpeed.Value);
            int step = Decimal.ToInt32(udServoStep.Value);

            // motors //

            if (IsKeyDown(Keys.NumPad7)) left += speed;
            if (IsKeyDown(Keys.NumPad4)) left -= speed;
            if (IsKeyDown(Keys.NumPad9)) right += speed;
            if (IsKeyDown(Keys.NumPad6)) right -= speed;
            if (IsKeyDown(Keys.NumPad5))
            {
                left += speed;
                right += speed;
            }
            if (IsKeyDown(Keys.NumPad2))
            {
                left -= speed;
                right -= speed;
            }
            if (IsKeyDown(Keys.NumPad1))
            {
                left -= speed;
                right += speed;
            }
            if (IsKeyDown(Keys.NumPad3))
            {
                left += speed;
                right -= speed;
            }
            if (left > speed) left = speed;
            if (left < -speed) left = -speed;
            if (right > speed) right = speed;
            if (right < -speed) right = -speed;
            robot.Drive(left, right);
            
            // left arm //

            robot.BeginUpdate();

            if (IsKeyDown(Keys.T)) robot.OffsetServoD("l_shoulder_rot", step);
            if (IsKeyDown(Keys.G)) robot.OffsetServoD("l_shoulder_rot", -step);
            if (IsKeyDown(Keys.R)) robot.OffsetServoD("l_shoulder_ud", step);
            if (IsKeyDown(Keys.F)) robot.OffsetServoD("l_shoulder_ud", -step);
            if (IsKeyDown(Keys.E)) robot.OffsetServoD("l_elbow", step);
            if (IsKeyDown(Keys.D)) robot.OffsetServoD("l_elbow", -step);
            if (IsKeyDown(Keys.W)) robot.OffsetServoD("l_wrist", step);
            if (IsKeyDown(Keys.S)) robot.OffsetServoD("l_wrist", -step);
            if (IsKeyDown(Keys.Q)) robot.OffsetServoD("l_hand", step);
            if (IsKeyDown(Keys.A)) robot.OffsetServoD("l_hand", -step);

            // right arm //

            if (IsKeyDown(Keys.U)) robot.OffsetServoD("r_shoulder_rot", step);
            if (IsKeyDown(Keys.H)) robot.OffsetServoD("r_shoulder_rot", -step);
            if (IsKeyDown(Keys.I)) robot.OffsetServoD("r_shoulder_ud", step);
            if (IsKeyDown(Keys.J)) robot.OffsetServoD("r_shoulder_ud", -step);
            if (IsKeyDown(Keys.O)) robot.OffsetServoD("r_elbow", step);
            if (IsKeyDown(Keys.K)) robot.OffsetServoD("r_elbow", -step);
            if (IsKeyDown(Keys.P)) robot.OffsetServoD("r_wrist", step);
            if (IsKeyDown(Keys.L)) robot.OffsetServoD("r_wrist", -step);
            if (IsKeyDown(Keys.Oem4)) robot.OffsetServoD("r_hand", step);
            if (IsKeyDown(Keys.Oem1)) robot.OffsetServoD("r_hand", -step);

            // head / base //

            if (IsKeyDown(Keys.Right)) robot.OffsetServoD("head_pan", step);
            if (IsKeyDown(Keys.Left)) robot.OffsetServoD("head_pan", -step);
            if (IsKeyDown(Keys.Up)) robot.OffsetServoD("head_tilt", step);
            if (IsKeyDown(Keys.Down)) robot.OffsetServoD("head_tilt", -step);
            if (IsKeyDown(Keys.X)) robot.OffsetServoD("base", step);
            if (IsKeyDown(Keys.Z)) robot.OffsetServoD("base", -step);

            robot.EndUpdate();
        }

        #endregion

        
        #region Script controls

        private void btnOpenScript_Click(object sender, EventArgs e)
        {
            OpenFileDialog d = new OpenFileDialog();
            d.InitialDirectory = Application.StartupPath;
            d.Filter = "Текстовые файлы (*.txt)|*.txt";
            d.CheckFileExists = true;
            if (DialogResult.OK == d.ShowDialog()) LoadScript(d.FileName);
        }

        /// <summary>
        /// Load a file "FileName" into rtScript and read sequence's names into listSeq
        /// </summary>
        /// <param name="FileName">path to file</param>
        private void LoadScript(string FileName)
        {
            if (System.IO.File.Exists(FileName))
            {
                btnStop.PerformClick();
                OpenFileName = FileName;
                rtScript.LoadFile(OpenFileName, RichTextBoxStreamType.PlainText);
                ReadSequences();
                log("script loaded (" + FileName + ")");
            }
            else log("script file not found (" + FileName + ")");
        }

        private void ReadSequences()
        {
            listSeq.Items.Clear();
            int i = 0;
            while (i < rtScript.Lines.Count())
            {
                if (rtScript.Lines[i].ToLower().StartsWith("seq=")) listSeq.Items.Add(rtScript.Lines[i].Substring(4));
                i++;
            }
        }

        private void btnSaveScript_Click(object sender, EventArgs e)
        {
            if (System.IO.File.Exists(OpenFileName))
            {
                rtScript.SaveFile(OpenFileName, RichTextBoxStreamType.PlainText);
                ReadSequences();
            }
            else btnSaveAs.PerformClick();
        }

        private void btnSaveAs_Click(object sender, EventArgs e)
        {
            SaveFileDialog d = new SaveFileDialog();
            d.InitialDirectory = Application.StartupPath;
            d.Filter = "Текстовые файлы (*.txt)|*.txt";
            d.AddExtension = true;
            d.DefaultExt = "txt";
            if (DialogResult.OK == d.ShowDialog()) rtScript.SaveFile(d.FileName, RichTextBoxStreamType.PlainText);
            ReadSequences();
        }

        private void listSeq_DoubleClick(object sender, EventArgs e)
        {
            if (listSeq.SelectedIndex >= 0) PlaySequence(listSeq.Items[listSeq.SelectedIndex].ToString());
        }

        private void PlaySequence(string name)
        {
            btnStop.PerformClick();
            CommandIndex = -1;
            while (CommandIndex < rtScript.Lines.Count() - 1)
            {
                CommandIndex++;
                if (rtScript.Lines[CommandIndex].StartsWith("seq=" + name, true, System.Globalization.CultureInfo.CurrentCulture)) break;
            }
            btnPlayPause.PerformClick();
        }

        private void btnPlayPause_Click(object sender, EventArgs e)
        {
            Playing = !Playing;
        }

        private void btnStop_Click(object sender, EventArgs e)
        {
            Playing = false;
            WaitUntil = 0;
            robot.SetServoSpeed(Decimal.ToInt32(udServoSpeed.Value)); // restore speed parameter //
            robot.EndUpdate();
            UpdateInfo();
        }

        private void btnStepPrevious_Click(object sender, EventArgs e)
        {
            CommandIndex--;
            if (CommandIndex >= 0) Exec(rtScript.Lines[CommandIndex].ToLower().Trim());
        }

        private void btnStepNext_Click(object sender, EventArgs e)
        {
            CommandIndex++;
            if (CommandIndex < rtScript.Lines.Count()) Exec(rtScript.Lines[CommandIndex].ToLower().Trim());
        }

        #endregion
        
        
        #region Command execution

        /// <summary>
        /// returns true if wait interval not elapsed
        /// </summary>
        private bool HaveToWait()
        {
            if (WaitUntil != 0)
            {
                if ((uint)Environment.TickCount >= WaitUntil) WaitUntil = 0;
            }
            return WaitUntil != 0;
        }

        /// <summary>
        /// reads a line from rtScript and executes it
        /// </summary>
        private void DoCommand()
        {
            if (HaveToWait()) return; // handle WAIT interval
            
            // fetch next line //

            CommandIndex++;
            if (rtScript.Lines.Count() <= CommandIndex)
            {
                btnStop.PerformClick();
                return;
            }

            string command_line = rtScript.Lines[CommandIndex].ToLower().Trim();
            rtScript.Focus();
            rtScript.Select(rtScript.GetFirstCharIndexFromLine(CommandIndex), command_line.Length);

            if (command_line.StartsWith("end")) // stop at the end of sequence
            {
                btnStop.PerformClick();
                return;
            }
            else if (command_line.StartsWith("//")) // skip comments
            {
                return;
            }

            try
            {
                WaitUntil = (uint)Environment.TickCount + (uint)Exec(command_line); // exec line and set WAIT variable
            }
            catch (Exception ex)
            {
                WaitUntil = 0;
                log(ex.Message);
            }
        }

        /// <summary>
        /// execute line of script
        /// </summary>
        /// <param name="script">a line of script</param>
        /// <returns>time to wait for this script to execute</returns>
        private int Exec(string script)
        {
            int i, index, position;
            int time = 0;

            robot.BeginUpdate();
            string[] commands = script.Split(' ');
            foreach (string command in commands)
            {
                try
                {
                    if (command.StartsWith("t"))
                    {
                        time = int.Parse(command.Substring(1));
                        robot.SetServoTime(time);
                    }
                    else if (command.StartsWith("w"))
                    {
                        time = int.Parse(command.Substring(1));
                    }
                    else if (command.StartsWith("da"))
                    {
                        robot.EnableAllServos(false);
                    }
                    else
                    {
                        i = command.IndexOf("=");
                        if (i > -1)
                        {
                            index = int.Parse(command.Substring(0, i));
                            position = int.Parse(command.Substring(i + 1));
                            robot.PositionServo(index, position);
                        }
                        else
                        {
                            index = int.Parse(command);
                            robot.EnableServo(index, false);
                        }
                    }
                }
                catch (Exception ex)
                {
                    throw new Exception("Script error at \"" + command + "\" at line \"" + script + "\". " + ex.Message);
                }
            }
            robot.EndUpdate();
            return time;
        }

        #endregion
        
        
        #region Camera

        private void btnCamera_Click(object sender, EventArgs e)
        {
            try
            {
                if (btnCamera.Checked)
                {
                    addLog("Включаю камеру ... ");
                    EnableCamera();
                    log("готово");
                }
                else
                {
                    addLog("Выключаю камеру ... ");
                    DisableCamera();
                    log("готово");
                }
            }
            catch (Exception ex)
            {
                log("Ошибка при работе с камерой. " + ex.Message);
                btnCamera.Checked = false;
            }

            // load cascade
            try
            {
                if (btnCamera.Checked) face_cascade = new HaarCascade(ExpandPath("%pp%\\haar\\haarcascade_frontalface_alt2.xml"));
                else face_cascade.Dispose();
            }
            catch (Exception ex)
            {
                log("Ошибка при загрузке/выгрузке каскада. " + ex.Message);
            }

            // load SURF tracker data
            try
            {
                if (btnCamera.Checked)
                {
                    chbMarker.Enabled = false;
                    if (System.IO.File.Exists(ExpandPath("%pp%\\image\\marker.png")))
                    {
                        surfParam = new MCvSURFParams(500, false);
                        marker = new Image<Gray, byte>(ExpandPath("%pp%\\image\\marker.png"));
                        markerFeatures = marker.ExtractSURF(ref surfParam);
                        chbMarker.Enabled = true;
                    }
                }
                else
                {
                    if (marker != null) marker.Dispose();
                }
            }
            catch (Exception ex)
            {
                log("Ошибка при загрузке/выгрузке маркера. " + ex.Message);
            }
        }

        public void EnableCamera()
        {
            DisableCamera();
            video_stream = new MJPEGStream(tbCameraAddress.Text);
            video_stream.NewFrame += new NewFrameEventHandler(NewFrame);
            video_stream.Start();
        }

        public void DisableCamera()
        {
            if (video_stream != null)
            {
                if (video_stream.IsRunning)
                {
                    video_stream.SignalToStop();
                    video_stream.WaitForStop();
                    video_stream = null;
                }
            }

            if (pic != null) pic.Image = null;
        }

        public void NewFrame(Object sender, NewFrameEventArgs e)
        {
            if (!working) DoCam((Bitmap)e.Frame.Clone());
        }

        private void DoCam(Bitmap bmp)
        {
            try
            {
                working = true;
                Stopwatch watch = Stopwatch.StartNew();

                MCvAvgComp[] faces = null;
                Point[] contour_points = null;
                Gray grayThreshold = new Gray(160);
                Gray grayThresholdLinking = new Gray(140);
                
                frame = new Image<Bgr, byte>(bmp);
                Image<Gray, Byte> gray = frame.Convert<Gray, Byte>();
                Image<Gray, Byte> edges = gray.Canny(grayThreshold, grayThresholdLinking);
                edges._Dilate(1);
                
                #region find face
                if (chbFace.Checked)
                {
                    faces = gray.DetectHaarCascade(face_cascade, 1.1, 3,
                       HAAR_DETECTION_TYPE.DO_CANNY_PRUNING | HAAR_DETECTION_TYPE.FIND_BIGGEST_OBJECT, new Size(50, 50))[0];
                    if (faces.Length > 0)
                    {
                        if (chbFollowFace.Checked)
                        {
                            int cx = (faces[0].rect.X + faces[0].rect.Width / 2 - gray.Width / 2) * 2 * 60 / gray.Width; // 60 is max speed
                            if (Math.Abs(cx) < 15) cx = 0;
                            robot.Drive(cx, -cx);
                        }
                    }
                    else if (chbFollowFace.Checked) robot.Drive(0, 0);
                }
                #endregion

                #region find contours
                if (chbContours.Checked)
                {
                    //MCvConvexityDefect[] defectarray;

                    using (MemStorage storage = new MemStorage())
                        for (Contour<Point> contours = edges.FindContours(CHAIN_APPROX_METHOD.CV_CHAIN_APPROX_SIMPLE, RETR_TYPE.CV_RETR_LIST, storage);
                             contours != null;
                             contours = contours.HNext)
                        {
                            Contour<Point> currentContour = contours.ApproxPoly(contours.Perimeter * 0.005, storage);
                            if (currentContour.Area > 250)
                            {
                                contour_points = currentContour.ToArray();
                                //Seq<MCvConvexityDefect> defects = currentContour.GetConvexityDefacts(storage, ORIENTATION.CV_COUNTER_CLOCKWISE);
                                //defectarray = defects.ToArray();
                                //foreach (MCvConvexityDefect defect in defectarray) ;
                            }
                        }
                }
                #endregion

                #region SURF tracker
                if (chbMarker.Checked)
                {
                    SURFTracker tracker = new SURFTracker(markerFeatures);
                    SURFFeature[] imageFeatures = gray.ExtractSURF(ref surfParam); // extract features
                    SURFTracker.MatchedSURFFeature[] matchedFeatures = tracker.MatchFeature(imageFeatures, 2, 20);
                    matchedFeatures = SURFTracker.VoteForUniqueness(matchedFeatures, 0.8);
                    matchedFeatures = SURFTracker.VoteForSizeAndOrientation(matchedFeatures, 1.5, 20);
                    HomographyMatrix homography = SURFTracker.GetHomographyMatrixFromMatchedFeatures(matchedFeatures);

                    if (homography != null)
                    {
                        Rectangle rect = marker.ROI;
                        PointF[] pts = new PointF[] { new PointF(rect.Left, rect.Bottom),
                            new PointF(rect.Right, rect.Bottom), new PointF(rect.Right, rect.Top), new PointF(rect.Left, rect.Top)};
                        homography.ProjectPoints(pts);
                        frame.DrawPolyline(Array.ConvertAll<PointF, Point>(pts, Point.Round), true, new Bgr(0, 0, 255), 1);
                    }

                    foreach (SURFTracker.MatchedSURFFeature matchedFeature in matchedFeatures)
                    {
                        frame.Draw(new CircleF(matchedFeature.ObservedFeature.Point.pt, 1), new Bgr(80, 80, 255), 1);
                        //frame.Draw(new LineSegment2DF(matchedFeature.SimilarFeatures[0].Feature.Point.pt, p), new Bgr(220, 240, 250), 1);
                    }
                }
                #endregion

                watch.Stop();

                // show frame //
                Graphics g = Graphics.FromImage(bmp);
                if (faces != null) { if (faces.Length > 0) g.DrawRectangle(new Pen(Color.Red, 1), faces[0].rect); }
                if (contour_points != null) g.DrawPolygon(new Pen(Color.DarkOrange, 2), contour_points);
                g.FillRectangle(new SolidBrush(Color.White), 0, 0, 40, 17);
                g.DrawString(watch.ElapsedMilliseconds.ToString() + "ms", new Font("Arial", 8), new SolidBrush(Color.Black), 2, 2);
                g.Dispose();
                pic.Image = bmp;
            }
            catch (Exception ex)
            {
                log("Ошибка захвата/обработки кадра. " + ex.Message);
            }
            finally
            {
                working = false;
            }
        }

        private void tbSaveFrame_Click(object sender, EventArgs e)
        {
            frame.Save(ExpandPath("%pp%\\image\\frame-" + DateTime.Now.ToString("yyyyMMdd-hhnnss") + ".png"));
        }

        #endregion

    }

}
